@@ -1,10 +1,10 @@
 ;; -*- no-byte-compile: t; lexical-binding: nil -*-
-(define-package "ocaml-eglot" "1.3.0"
+(define-package "ocaml-eglot" "20251022.937"
   "An OCaml companion for Eglot."
   '((emacs "29.1"))
   :url "https://github.com/tarides/ocaml-eglot"
-  :commit "69d4951ee449fc8981134258cada85dd59412855"
-  :revdesc "69d4951ee449"
+  :commit "3259c506dd6fe85e4b61abfa8a68706f4850f9a1"
+  :revdesc "3259c506dd6f"
   :keywords '("ocaml" "languages")
   :authors '(("Xavier Van de Woestyne" . "xaviervdw@gmail.com"))
   :maintainers '(("Xavier Van de Woestyne" . "xaviervdw@gmail.com")))
