@@ -13,8 +13,8 @@
 ;; Keywords: ocaml languages
 ;; URL: https://github.com/tarides/ocaml-eglot
 ;; Package-Requires: ((emacs "29.1"))
-;; Package-Version: 1.3.0
-;; Package-Revision: 69d4951ee449
+;; Package-Version: 20251022.937
+;; Package-Revision: 3259c506dd6f
 ;; SPDX-License-Identifier: MIT
 
 ;;; Commentary:
@@ -41,7 +41,6 @@
 (require 'ocaml-eglot-req)
 (require 'ocaml-eglot-type-enclosing)
 (require 'ocaml-eglot-xref)
-(require 'ocaml-eglot-objinfo)
 (require 'eglot)
 
 (declare-function flycheck-next-error "ext:flycheck")
@@ -92,6 +91,14 @@ Otherwise, `merlin-construct' only includes constructors."
                        (featurep 'flycheck-eglot))
               '((const :tag "Use Flycheck" flycheck)))))
 
+(defcustom ocaml-eglot-objinfo-flags
+  (list "-shape" "-index" "-decls" "-uid-deps")
+  "Flags passed to `ocamlobjinfo'."
+  :type '(set
+          (const "-shape")
+          (const "-index")
+          (const "-decls")
+          (const "-uid-deps")))
 
 ;;; Faces
 
@@ -146,32 +153,107 @@ Otherwise, `merlin-construct' only includes constructors."
     (_ (ocaml-eglot--invalid-syntax-checker))))
 
 ;; Jump to definition
-;; The following commands are based primarily on `xref' and are
-;; here in correspondence with the `merlin` mode.
 
-(defalias 'ocaml-eglot-locate #'xref-find-definitions)
-(defalias 'ocaml-eglot-find-definition #'xref-find-definitions)
+(defun ocaml-eglot--find (strategy query)
+  "Find the definition using QUERY and jump to it using STRATEGY."
+  (let* ((query-result (funcall query))
+         (result (ocaml-eglot-util--vec-first-or-nil query-result)))
+    (if result
+        (let* ((uri (cl-getf result :uri))
+               (range (cl-getf result :range))
+               (file (ocaml-eglot-util--uri-to-path uri)))
+          (ocaml-eglot-util--visit-file strategy (buffer-file-name) file range))
+      (eglot--error "Not in environment"))))
+
+(defun ocaml-eglot--find-definition (strategy)
+  "Find the definition at point and jump to it using STRATEGY."
+  (ocaml-eglot--find strategy #'ocaml-eglot-req--definition))
+
+(defun ocaml-eglot-find-definition ()
+  "Find the definition identifier at point."
+  (interactive)
+  (ocaml-eglot--find-definition ocaml-eglot-open-window-strategy))
+
+(defun ocaml-eglot-find-definition-in-new-window ()
+  "Find the definition of the identifier at point and show it in a new window."
+  (interactive)
+  (ocaml-eglot--find-definition 'new))
+
+(defun ocaml-eglot-find-definition-in-current-window ()
+  "Find the definition of the identifier and show it in the current window."
+  (interactive)
+  (ocaml-eglot--find-definition 'current))
+
+(defun ocaml-eglot--identifier-query (identifier kind)
+  "Build the locate IDENTIFIER query based on  KIND."
+  (lambda () (let* ((result (ocaml-eglot-req--locate-ident identifier kind))
+                    (result-value (ocaml-eglot-util--merlin-call-result result))
+                    (loc (ocaml-eglot-util--merlin-location-to-lsp result-value)))
+               (make-vector 1 loc))))
+
+(defun ocaml-eglot--find-identifier-definition (identifier strategy)
+  "Find the definition of IDENTIFIER jump to it using STRATEGY."
+  (ocaml-eglot--find
+   strategy (ocaml-eglot--identifier-query identifier 'implementation)))
+
+(defun ocaml-eglot-find-identifier-definition (identifier)
+  "Find the definition of the given IDENTIFIER."
+  (interactive "s> ")
+  (ocaml-eglot--find-identifier-definition
+   identifier ocaml-eglot-open-window-strategy))
+
+(defun ocaml-eglot-find-identifier-definition-in-new-window (identifier)
+  "Find the definition of the IDENTIFIER and show it in a new window."
+  (interactive "s> ")
+  (ocaml-eglot--find-identifier-definition identifier 'new))
+
+(defun ocaml-eglot-find-identifier-definition-in-current-window (identifier)
+  "Find the definition of the IDENTIFIER and show it in the current window."
+  (interactive "s> ")
+  (ocaml-eglot--find-identifier-definition identifier 'current))
+
+;; Jump to declaration
+
+(defun ocaml-eglot--find-declaration (strategy)
+  "Find the declaration of the identifier at point and jump to it using STRATEGY."
+  (ocaml-eglot--find strategy #'ocaml-eglot-req--declaration))
 
 (defun ocaml-eglot-find-declaration ()
-  "Find the declaration of identifier at point."
+  "Find the declaration of the identifier at point."
   (interactive)
-  (let ((ocaml-eglot-locate-preference 'mli)
-        (xref-prompt-for-identifier nil))
-    (call-interactively #'xref-find-definitions)))
+  (ocaml-eglot--find-declaration ocaml-eglot-open-window-strategy))
 
-(defun ocaml-eglot-find-identifier-definition ()
-  "Find the definition of identifier at point."
+(defun ocaml-eglot-find-declaration-in-new-window ()
+  "Find the declaration of the identifier at point and show it in a new window."
   (interactive)
-  (let ((ocaml-eglot-locate-preference 'ml)
-        (xref-prompt-for-identifier t))
-    (call-interactively #'xref-find-definitions)))
+  (ocaml-eglot--find-declaration 'new))
 
-(defun ocaml-eglot-find-identifier-declaration ()
-  "Find the declaration of identifier at point."
+(defun ocaml-eglot-find-declaration-in-current-window ()
+  "Find the declaration of the identifier at point.
+Show it the current window."
   (interactive)
-  (let ((ocaml-eglot-locate-preference 'mli)
-        (xref-prompt-for-identifier t))
-    (call-interactively #'xref-find-definitions)))
+  (ocaml-eglot--find-declaration 'current))
+
+(defun ocaml-eglot--find-identifier-declaration (identifier strategy)
+  "Find the declaration of IDENTIFIER jump to it using STRATEGY."
+  (ocaml-eglot--find
+   strategy (ocaml-eglot--identifier-query identifier 'interface)))
+
+(defun ocaml-eglot-find-identifier-declaration (identifier)
+  "Find the declaration of the given IDENTIFIER."
+  (interactive "s> ")
+  (ocaml-eglot--find-identifier-declaration
+   identifier ocaml-eglot-open-window-strategy))
+
+(defun ocaml-eglot-find-identifier-declaration-in-new-window (identifier)
+  "Find the declaration of the given IDENTIFIER and show it in a new window."
+  (interactive "s> ")
+  (ocaml-eglot--find-identifier-declaration identifier 'new))
+
+(defun ocaml-eglot-find-identifier-declaration-in-current-window (identifier)
+  "Find the declaration of the given IDENTIFIER and show it in the current window."
+  (interactive "s> ")
+  (ocaml-eglot--find-identifier-declaration identifier 'current))
 
 ;; Jump type declaration of expression
 
@@ -189,7 +271,7 @@ Otherwise, `merlin-construct' only includes constructors."
 (defun ocaml-eglot-find-type-definition ()
   "Find the definition of the type of the expression at point."
   (interactive)
-  (ocaml-eglot--find-type-definition 'current))
+  (ocaml-eglot--find-type-definition ocaml-eglot-open-window-strategy))
 
 (defun ocaml-eglot-find-type-definition-in-new-window ()
   "Find the definition of the type of the expression at point.
@@ -246,45 +328,10 @@ If optional IN-OTHER-WINDOW is non-nil, find the file in another window."
         (find-file-other-window file)
       (find-file file))))
 
-(defun ocaml-eglot-find-identifier-in-alternate-file (&optional in-other-window)
-  "Find the declaration of the identifier at point in the alternative file.
-
-If optional IN-OTHER-WINDOW is non-nil, find the declaration in another window."
-  (interactive "P")
-  (let* ((identifier (xref-backend-identifier-at-point (xref-find-backend)))
-         (identifier-str (progn (substring-no-properties identifier)))
-         (alternate-buffer
-          (progn
-            (ocaml-eglot-req--server-capable-or-lose :experimental :ocamllsp :handleSwitchImplIntf)
-            (when-let* ((current-uri (ocaml-eglot-util--current-uri))
-                        (uri (ocaml-eglot--find-alternate-file current-uri))
-                        (file (ocaml-eglot-util--uri-to-path uri)))
-              (or (get-file-buffer file) (find-file-noselect file))))))
-    (xref-push-marker-stack)
-    ;; Find, or create, the buffer for the alternate file.
-    (condition-case error
-        (with-current-buffer alternate-buffer
-          ;; We go to the end of buffer to work around https://github.com/ocaml/ocaml-lsp/issues/1586
-          (goto-char (point-max))
-          (if in-other-window
-              (xref-find-definitions-other-window identifier-str)
-            (xref-find-definitions identifier-str))
-          ;; Restore `xref--history'. `xref-find-definitions' will put
-          ;; a marker at the end of the alternate file, but we want to
-          ;; jump back to the non-alternative file rather.
-          (setq xref--history `(,(cdr (car xref--history)) . ,(cdr xref--history))))
-      (user-error
-       ;; For some reason the LSP sometimes fails find the
-       ;; definition/declaration of values. Type
-       ;; declaration/definitions do not seem suffer from this issue.
-       ;; When this happens, `xref-find-definitions' throws an
-       ;; `user-error' that we catch here.
-       (user-error "Couldn't find %s in `%s', error message from LSP: %s" identifier (buffer-name alternate-buffer) (cadr error))))))
-
 ;; Hook when visiting new interface file
 
 (defun ocaml-eglot--file-hook ()
-  "Hook to try to generate interface on visiting new files."
+  "Hook to try to generate interface on visiting new files.."
   (when (and
          (ocaml-eglot-util--on-interface)
          (= (buffer-size) 0))
@@ -389,16 +436,12 @@ If there is no available holes, it returns the first one of HOLES."
 
 (defun ocaml-eglot--phrase (direction)
   "Move to the next or previous phrase using DIRECTION."
-  (if (ocaml-eglot-req--server-capable :experimental :ocamllsp :handlePhrase)
-      (let ((result (ocaml-eglot-req--phrase direction)))
-        (when result
-          (ocaml-eglot-util--jump-to result)))
-    (let* ((result (ocaml-eglot-req--phrase-legacy direction))
-           (json-result (ocaml-eglot-util--merlin-call-result result))
-           (pos (cl-getf json-result :pos)))
-      (when pos
-        (let ((target (ocaml-eglot-util--pos-to-point pos)))
-          (ocaml-eglot-util--goto-char target))))))
+  (let* ((result (ocaml-eglot-req--phrase direction))
+         (json-result (ocaml-eglot-util--merlin-call-result result))
+         (pos (cl-getf json-result :pos)))
+    (when pos
+      (let ((target (ocaml-eglot-util--pos-to-point pos)))
+        (ocaml-eglot-util--goto-char target)))))
 
 (defun ocaml-eglot-phrase-next ()
   "Go to the beginning of the next phrase."
@@ -499,9 +542,27 @@ The universal prefix argument can be used to change the maximum number
 of results (LIMIT)."
   (interactive "sSearch query: \np")
   (ocaml-eglot--search-def-or-decl
-   (lambda (identifier)
-     (let ((ocaml-eglot-locate-preference 'ml))
-       (xref-find-definitions identifier)))
+   #'ocaml-eglot-find-identifier-definition
+   query
+   limit))
+
+(defun ocaml-eglot-search-definition-in-current-window (query &optional limit)
+  "Search a definition using a QUERY (type or polarity) in the current window.
+The universal prefix argument can be used to change the maximum number
+of results (LIMIT)."
+  (interactive "sSearch query: \np")
+  (ocaml-eglot--search-def-or-decl
+   #'ocaml-eglot-find-identifier-definition-in-current-window
+   query
+   limit))
+
+(defun ocaml-eglot-search-definition-in-new-window (query &optional limit)
+  "Search a definition using a QUERY (type or polarity) in a new window.
+The universal prefix argument can be used to change the maximum number
+of results (LIMIT)."
+  (interactive "sSearch query: \np")
+  (ocaml-eglot--search-def-or-decl
+   #'ocaml-eglot-find-identifier-definition-in-new-window
    query
    limit))
 
@@ -511,9 +572,27 @@ The universal prefix argument can be used to change the maximum number
 of results (LIMIT)."
   (interactive "sSearch query: \np")
   (ocaml-eglot--search-def-or-decl
-   (lambda (identifier)
-     (let ((ocaml-eglot-locate-preference 'mli))
-       (xref-find-definitions identifier)))
+   #'ocaml-eglot-find-identifier-declaration
+   query
+   limit))
+
+(defun ocaml-eglot-search-declaration-in-current-window (query &optional limit)
+  "Search a declaration using a QUERY (type or polarity) in the current window.
+The universal prefix argument can be used to change the maximum number
+of results (LIMIT)."
+  (interactive "sSearch query: \np")
+  (ocaml-eglot--search-def-or-decl
+   #'ocaml-eglot-find-identifier-declaration-in-current-window
+   query
+   limit))
+
+(defun ocaml-eglot-search-declaration-in-new-window (query &optional limit)
+  "Search a declaration using a QUERY (type or polarity) in a new window.
+The universal prefix argument can be used to change the maximum number
+of results (LIMIT)."
+  (interactive "sSearch query: \np")
+  (ocaml-eglot--search-def-or-decl
+   #'ocaml-eglot-find-identifier-declaration-in-new-window
    query
    limit))
 
@@ -580,13 +659,9 @@ It use the ARG to use local values or not."
 (defun ocaml-eglot-type-expression (expression)
   "Prompt the user for expression EXPRESSION and print its type."
   (interactive "sExpression: ")
-  (if (ocaml-eglot-req--server-capable
-       :experimental :ocamllsp :handleTypeExpression)
-      (let ((type-expr (ocaml-eglot-req--type-expression expression)))
-        (ocaml-eglot-type-enclosing--display type-expr nil))
-    (let* ((result (ocaml-eglot-req--type-expression-legacy expression))
-           (type-expr (ocaml-eglot-util--merlin-call-result result)))
-      (ocaml-eglot-type-enclosing--display type-expr nil))))
+  (let* ((result (ocaml-eglot-req--type-expression expression))
+         (type-expr (ocaml-eglot-util--merlin-call-result result)))
+    (ocaml-eglot-type-enclosing--display type-expr nil)))
 
 (defun ocaml-eglot-type-enclosing (&optional prefix)
   "Print the type of the expression under point (or of the region, if it exists).
@@ -598,11 +673,6 @@ and print its type."
       (call-interactively #'ocaml-eglot-type-expression)
     (ocaml-eglot-type-enclosing--call)))
 
-(defun ocaml-eglot-type-annotate ()
-  "Annotate the type of the expression at point."
-  (interactive)
-  (ocaml-eglot-type-enclosing)
-  (ocaml-eglot-type-enclosing-annotate))
 
 ;; Case Analysis
 
@@ -642,15 +712,13 @@ and print its type."
 
 (cl-defmethod eglot-client-capabilities :around (_)
   "Add client capabilities to Eglot for OCaml LSP server."
-(let* ((capabilities (copy-tree (cl-call-next-method)))
-         (experimental-capabilities
-          (if-let* ((previous (cl-getf capabilities :experimental)))
-              (copy-hash-table previous)
-            (make-hash-table)))
+  (let* ((capabilities (copy-tree (cl-call-next-method)))
+         (experimental-capabilities (cl-getf capabilities :experimental))
+         (previous (or experimental-capabilities eglot--{}))
          (commands (append (apply #'vector ocaml-eglot-client-capabilities) nil)))
     (dolist (key commands)
-      (puthash key t experimental-capabilities))
-    (setq capabilities (plist-put capabilities :experimental experimental-capabilities))))
+      (puthash key t previous))
+    (setq capabilities (plist-put capabilities :experimental previous))))
 
 ;; A command can be executed by the server or by the client. The
 ;; following code analyses the command. If it's a command which must
@@ -722,5 +790,40 @@ OCaml Eglot provides standard implementations of the various custom-requests
   :group 'ocaml-eglot
   (if ocaml-eglot (ocaml-eglot-setup) (ocaml-eglot-clean)))
 
+;;; Ocamlobjinfo mode
+
+;; Allows known compilation artefacts to be displayed via the
+;; `ocamlobjinfo' binary
+
+(define-derived-mode ocaml-eglot-objinfo-mode special-mode "OCaml-obj-info"
+  "Mode for displaying ocamlobjinfo output."
+  (setq buffer-read-only t
+        buffer-offer-save nil
+        buffer-file-name nil
+        auto-save-default nil
+        make-backup-files nil
+        create-lockfiles nil)
+  (define-key ocaml-eglot-objinfo-mode-map (kbd "q") #'quit-window))
+
+;;;###autoload
+(defun ocaml-eglot-objinfo-handler ()
+  "Display the result of `ocamlobjinfo` instead of file contents."
+  (when (and buffer-file-name
+             (ocaml-eglot-util--is-artifact buffer-file-name))
+    (let ((file buffer-file-name)
+          (inhibit-read-only t))
+      (setq buffer-read-only nil)
+      (erase-buffer)
+      (if (executable-find "ocamlobjinfo")
+          (let ((args (append ocaml-eglot-objinfo-flags (list file))))
+            (apply #'call-process "ocamlobjinfo" nil t nil args))
+        (insert "`ocamlobjinfo' not found in PATH"))
+      (goto-char (point-min))
+      (ocaml-eglot-objinfo-mode))))
+
+
+;;;###autoload
+(add-hook 'find-file-hook #'ocaml-eglot-objinfo-handler)
+
 (provide 'ocaml-eglot)
 ;;; ocaml-eglot.el ends here
